/*
 *  dmidid.c  Linux / ALSA 0.9 DMIDI client
 *  Version: 0.4
 *  Homepage: http://www.dmidi.org
 *
 *  Copyright (C) 2003 Phil Kerr
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <alsa/asoundlib.h>

#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <pthread.h>
#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <net/if_packet.h>
#include <linux/if_ether.h>
#include <net/if_arp.h>
#include <sys/ioctl.h>

#define BUFSIZE 1500

pthread_t dmidiInThread; 

unsigned char* dmidi_output [BUFSIZE];

// Default receive node ID
unsigned char rxn1 = 0xff;
unsigned char rxn2 = 0xff;
unsigned char rxn3 = 0xff;
unsigned char rxn4 = 0xff;

// Default transmit node ID
unsigned char txn1 = 0xff;
unsigned char txn2 = 0xff;
unsigned char txn3 = 0xff;
unsigned char txn4 = 0xff;

int broadcastFlag = 0;

unsigned char myMACAddr [6];

int queue_id, port_in_id, port_out_id;
snd_seq_t *seq_handle;

//----------------------------------------------------------------------------//
// Prototype definitiions

void alsaseqout (); 
snd_seq_t *open_tx ();
void dmidi_tx (snd_seq_t *seq_handle);
void open_eth_ports ();

//----------------------------------------------------------------------------//
// SCHED_FIFO realtime priority function

int set_realtime_priority (void)
{
    struct sched_param schp;

    memset (&schp, 0, sizeof (schp));
    schp.sched_priority = sched_get_priority_max (SCHED_FIFO);

    if (sched_setscheduler (0, SCHED_FIFO, &schp) != 0) {
            perror ("sched_setscheduler");
            return -1;
    }
            return 0;
}


//----------------------------------------------------------------------------//
// Send DMIDI data function.  This outputs the notes via DMIDI broadcast

void send_dmidi (int len, unsigned char * dmidi_output,
                 unsigned char n1, unsigned char n2, 
                 unsigned char n3, unsigned char n4)
{
    int res, xlen, loop = 0;
    struct sockaddr sa;
    struct sockaddr * sap = &sa;
    struct ifreq ifr;
    unsigned char buffer [1500];
    unsigned char ebroadcast [1];

    // create socket
    int s = socket (AF_INET, SOCK_PACKET, htons (ETH_P_ALL));

    if (s < 0) {
      perror ("socket");
      exit (1);
    }

    if (sizeof (sa) < sizeof (struct sockaddr)) {
        printf ("illegal size %d, should be %d\n",
            sizeof (sa), sizeof (struct sockaddr));
        exit (1);
    }

    // socket setup
    xlen = sizeof (sa);
    bzero (sap, xlen);
    sa.sa_family = AF_INET;
    strcpy (sa.sa_data, "eth0");

    bzero (&ifr, sizeof(ifr));
    strcpy (ifr.ifr_name, "eth0");
    if (ioctl (s, SIOCGIFHWADDR, &ifr) < 0) {
      perror ("ioctl(SIOCGIFHWADDR)");
      exit (1);
    }

    memcpy (myMACAddr, ifr.ifr_hwaddr.sa_data, 6);

    ebroadcast [0] = 0xff;
    bzero (buffer, BUFSIZE);

//    memcpy (buffer, ebroadcast, 6); 
    buffer [0] = 0xff;
    buffer [1] = 0xff;
    buffer [2] = 0xff;
    buffer [3] = 0xff;
    buffer [4] = 0xff;
    buffer [5] = 0xff;

    memcpy (buffer + 6, ifr.ifr_hwaddr.sa_data, 6); 

    // dmidi ethernet type
    buffer [12] = 0x11;
    buffer [13] = 0x99;

    // add node ID's to output packet
    buffer [14] = n1;
    buffer [15] = n2;
    buffer [16] = n3;
    buffer [17] = n4;
 
    buffer [18] = 0xff;
    buffer [19] = 0xff;
    buffer [20] = 0xff;
    buffer [21] = 0xff;
    buffer [22] = 0xff;
    buffer [23] = 0xff;

    if (len <= 44) buffer [24] = len;
    else buffer [24] = 0xff;

    memcpy (buffer + 25, dmidi_output, len); // dmidi data 

//for (loop = 0 ; loop < 25 + len ; loop++) printf ("%x ",buffer[loop]);
//printf ("\n");

    // send data to socket
    res = sendto (s, buffer, (25 + len), 0, sap, sizeof (sa));

    // close socket
    close (s);
}

//----------------------------------------------------------------------------//
//

void dmidi_tx (snd_seq_t *seq_handle) {

    unsigned char dmidi_data [BUFSIZE];
    snd_seq_event_t *ev;

int sysexlen;
unsigned char *sysexdata;


    do {
        snd_seq_event_input (seq_handle, &ev);

        switch (ev->type) {

        case SND_SEQ_EVENT_CONTROLLER: 

            dmidi_data [0] = (0xb0 + ev -> data.control.channel);
            dmidi_data [1] = ev -> data.control.param;
            dmidi_data [2] = ev -> data.control.value;

            send_dmidi (3, dmidi_data, txn1, txn2, txn3, txn4);
            break;

        case SND_SEQ_EVENT_PITCHBEND:

            dmidi_data [0] = (0xe0 + (ev -> data.control.channel));
            dmidi_data [1] = ((ev -> data.control.value) & 0x7f);
            dmidi_data [2] = (0x40 +  ((ev -> data.control.value) >> 7));

            send_dmidi (3, dmidi_data, txn1, txn2, txn3, txn4);
            break;

        case SND_SEQ_EVENT_NOTEON:

            dmidi_data [0] = (0x90 + ev -> data.control.channel);
            dmidi_data [1] = ev -> data.note.note;
            dmidi_data [2] = ev -> data.note.velocity;

            send_dmidi (3, dmidi_data, txn1, txn2, txn3, txn4);
            break;        

        case SND_SEQ_EVENT_NOTEOFF: 
            dmidi_data [0] = (0x80 + ev -> data.control.channel);
            dmidi_data [1] = ev -> data.note.note;
            dmidi_data [2] = ev -> data.note.velocity;

            send_dmidi (3, dmidi_data, txn1, txn2, txn3, txn4);
            break;        

        case SND_SEQ_EVENT_SYSEX: 
            sysexlen = ev -> data.ext.len;
            sysexdata = ev -> data.ext.ptr;

            send_dmidi (sysexlen, sysexdata, txn1, txn2, txn3, txn4);
            break;        
        }

    snd_seq_free_event (ev);

    } while (snd_seq_event_input_pending (seq_handle, 0) > 0);
}

//----------------------------------------------------------------------------//
//  Open ALSA device

snd_seq_t *open_seq() {

    char alsaportname [100];
    char txportlabel [100];
    int portid;

    if (snd_seq_open (&seq_handle, "default", SND_SEQ_OPEN_DUPLEX, 0) < 0) {
        fprintf (stderr, "Error opening ALSA sequencer.\n");
        exit(1);
    }

    sprintf (alsaportname, "DMIDI - Receive: [%02x:%02x:%02x:%02x]", rxn1, rxn2, rxn3, rxn4);

    snd_seq_set_client_name (seq_handle, "DMIDI");

    if ( (port_out_id = snd_seq_create_simple_port(seq_handle, alsaportname,
            SND_SEQ_PORT_CAP_READ|SND_SEQ_PORT_CAP_SUBS_READ,
            SND_SEQ_PORT_TYPE_APPLICATION)) < 0) {
    fprintf (stderr, "Error creating sequencer port.\n");
    }


    sprintf (txportlabel, "DMIDI - Transmit [%02x:%02x:%02x:%02x]", txn1, txn2, txn3, txn4);

    snd_seq_set_client_name (seq_handle, "DMIDI");

    if ((portid = snd_seq_create_simple_port (seq_handle, txportlabel,
            SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE,
            SND_SEQ_PORT_TYPE_APPLICATION)) < 0) {
    fprintf (stderr, "Error creating sequencer port.\n");
    }

    return (seq_handle);
}

//----------------------------------------------------------------------------//
//

void* dmidi_rx (void* seq_handle) {

    int npfd, loop;
    struct pollfd *pfd;
    struct ifreq ifr;

    unsigned int nb;
    unsigned char inbuf [BUFSIZE];
    unsigned char alsadata [BUFSIZE];
    unsigned char tmpmac [6];

    int res = 0, xlen;
    struct sockaddr sa;
    struct sockaddr * sap = &sa;

    // create raw ethernet socket
    int s = socket (AF_INET, SOCK_PACKET, htons (ETH_P_ALL));

    if (s < 0) {
      perror ("socket");
      exit (1);
    }

    if (sizeof (sa) < sizeof (struct sockaddr)) {
        printf ("illegal size %d, should be %d\n",
            sizeof (sa), sizeof (struct sockaddr));
        exit (1);
    }

    bzero (&ifr, sizeof(ifr));
    strcpy (ifr.ifr_name, "eth0");
    if (ioctl (s, SIOCGIFHWADDR, &ifr) < 0) {
      perror ("ioctl(SIOCGIFHWADDR)");
      exit (1);
    }

    memcpy (myMACAddr, ifr.ifr_hwaddr.sa_data, 6);

    // socket setup
    xlen = sizeof (sa);

    queue_id = snd_seq_alloc_queue (seq_handle);
    snd_seq_set_client_pool_output (seq_handle, 1);

    snd_seq_start_queue (seq_handle, queue_id, NULL);
    snd_seq_drain_output (seq_handle);
    npfd = snd_seq_poll_descriptors_count (seq_handle, POLLIN);
    pfd = (struct pollfd *) alloca (npfd * sizeof(struct pollfd));
    snd_seq_poll_descriptors (seq_handle, pfd, npfd, POLLIN);


    while (1) {
        memset (&inbuf, 0, sizeof (inbuf));
        memset (&tmpmac, 0, sizeof (tmpmac));

        res = recvfrom (s, inbuf, BUFSIZE, 0, sap, &xlen);
        if (res < 0) {
            perror ("recvfrom");
            exit (1);
        }

        memcpy (tmpmac, inbuf + 6, 6);

        if (memcmp (myMACAddr, tmpmac, 6) != 0) {

            if ( ((inbuf [12] == 0x11) && (inbuf [13] == 0x99) &&  
               (inbuf [14] == rxn1) && (inbuf [15] == rxn2) && 
               (inbuf [16] == rxn3) && (inbuf [17] == rxn4)) ||

               ((inbuf [12] == 0x11) && (inbuf [13] == 0x99) &&  
               (inbuf [14] == 0xff) && (inbuf [15] == 0xff) && 
               (inbuf [16] == 0xff) && (inbuf [17] == 0xff) && broadcastFlag == 1) ) {

                    memcpy (alsadata, inbuf + 25, res - 25); 
                    alsaseqout (nb - 4, alsadata);
            }
        }  
    }
}

//----------------------------------------------------------------------------//
//  Dumps DMIDI data to ALSA port
                           
void alsaseqout (unsigned int len, unsigned char* data) 
{
    snd_seq_event_t ev;

    snd_seq_ev_clear (&ev);

    if ((data [0]  & 0xf0) == 0x80) snd_seq_ev_set_noteoff (&ev, 0, data[1], data[2]);
    if ((data [0]  & 0xf0) == 0x90) snd_seq_ev_set_noteon (&ev, 0, data[1], data[2]);
    if ((data [0]  & 0xf0) == 0xb0) snd_seq_ev_set_controller (&ev, data [0] - 0xb0, data[1], data[2]);
    if ((data [0]  & 0xf0) == 0xe0) snd_seq_ev_set_pitchbend (&ev, data[1], data[2]);
    if ((data [0]  & 0xf0) == 0xf0) snd_seq_ev_set_sysex (&ev, *data, &len);


    snd_seq_ev_set_source (&ev, port_out_id);
    snd_seq_ev_set_subs (&ev);
    snd_seq_event_output_direct (seq_handle, &ev);

    snd_seq_ev_clear (&ev);
    ev.type = SND_SEQ_EVENT_ECHO; 
    snd_seq_ev_set_dest (&ev, snd_seq_client_id (seq_handle), port_in_id);
    snd_seq_event_output_direct (seq_handle, &ev);
}  

//----------------------------------------------------------------------------//

int main (int argc, char *argv[]) {

    snd_seq_t *seq_handle;
    int npfd;
    struct pollfd *pfd;
    int status, i;    
    unsigned char *noderx;
    unsigned char *temprxn1;
    unsigned char *temprxn2;
    unsigned char *temprxn3;
    unsigned char *temprxn4;

    unsigned char *nodetx;
    unsigned char *temptxn1;
    unsigned char *temptxn2;
    unsigned char *temptxn3;
    unsigned char *temptxn4;

    char *end;

    const char delimiters[] = ":";

    for (i = 1; i < argc; i++)
    {
        if (argv [i][0] == '-')
	{
            switch (argv [i][1])
	    {
	    case 'v':
	        printf ("\n\tdmidid - Linux Client - 0.4 (ALSA 0.9): www.dmidi.org:\n\n");;
                exit (1);
            break;

	    case 'h':
            // todo
            break;

	    case 'b':
                broadcastFlag = 1;
            break;

	    case 'r':
	        if (i + 1 < argc) noderx = argv [++i];

                if ((strlen (noderx) == 11)) {

                    temprxn1 = strtok (noderx, delimiters);
                    temprxn2 = strtok (NULL, delimiters);
                    temprxn3 = strtok (NULL, delimiters);
                    temprxn4 = strtok (NULL, delimiters);

                    rxn1 = (char) strtol (temprxn1, &end, 16);
                    rxn2 = (char) strtol (temprxn2, &end, 16);
                    rxn3 = (char) strtol (temprxn3, &end, 16);
                    rxn4 = (char) strtol (temprxn4, &end, 16);
                }
            break;

	    case 't':
	        if (i + 1 < argc) nodetx = argv [++i];

                if ((strlen (nodetx) == 11)) {

                    temptxn1 = strtok (nodetx, delimiters);
                    temptxn2 = strtok (NULL, delimiters);
                    temptxn3 = strtok (NULL, delimiters);
                    temptxn4 = strtok (NULL, delimiters);

                    txn1 = (char) strtol (temptxn1, &end, 16);
                    txn2 = (char) strtol (temptxn2, &end, 16);
                    txn3 = (char) strtol (temptxn3, &end, 16);
                    txn4 = (char) strtol (temptxn4, &end, 16);
                }
            break;

	    }
	}
    }

    // Set RT priority
    set_realtime_priority ();

    // Print greeting/node numbers
    fprintf (stdout, "\ndmidid - Linux client - 0.4 (ALSA 0.9)\n");
    fprintf (stdout, "Transmit Node ID: [%02x:%02x:%02x:%02x]\n", txn1, txn2, txn3, txn4);
    fprintf (stdout, "Receive  Node ID: [%02x:%02x:%02x:%02x]\n", rxn1, rxn2, rxn3, rxn4);
    if (broadcastFlag == 1)
    fprintf (stdout, "Receive  Node ID: [ff:ff:ff:ff]\n");

    // Open ALSA writeable port for DMIDI transmission
    seq_handle = open_seq ();

    // Create DMIDI receive thread and check status
    status = pthread_create (&dmidiInThread, NULL, dmidi_rx, seq_handle);

    if (status == -1) {
        printf ("Error: unable to create DMIDI input thread.\n");
    } 

    // Setup ALSA event loop
    npfd = snd_seq_poll_descriptors_count (seq_handle, POLLIN);
    pfd = (struct pollfd *) alloca(npfd * sizeof(struct pollfd));
    snd_seq_poll_descriptors (seq_handle, pfd, npfd, POLLIN);

    // Main DMIDI broadcast loop ALSA poll
    while (1) {
        if (poll (pfd, npfd, 100000) > 0) {
            dmidi_tx (seq_handle);
        }  
    }
}

//  End
//----------------------------------------------------------------------------//
